# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# The version is stored in _version.py as recommended here:
# http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
from ._version import __version__, __version_info__
import os,cffi,glob

# Also let coord.version show the version.
version = __version__

# Set module level attributes for the include directory and the library file name.
coord_dir = os.path.dirname(__file__)
include_dir = os.path.join(coord_dir,'include')

lib_file = os.path.join(coord_dir,'_coord.so')
# Some installation (e.g. Travis with python 3.x) name this e.g. _coord.cpython-34m.so,
# so if the normal name doesn't exist, look for something else.
if not os.path.exists(lib_file): # pragma: no cover
    alt_files = glob.glob(os.path.join(coord_dir,'_coord*.so'))
    if len(alt_files) == 0:
        raise IOError("No file '_coord.so' found in %s"%coord_dir)
    if len(alt_files) > 1:
        raise IOError("Multiple files '_coord*.so' found in %s: %s"%(coord_dir,alt_files))
    lib_file = alt_files[0]

# Load the C functions with cffi
_ffi = cffi.FFI()
for file_name in glob.glob(os.path.join(include_dir,'*_C.h')):
    with open(file_name) as f:
        _ffi.cdef(f.read())
_lib = _ffi.dlopen(lib_file)

# Explicitly import things that we want to be in the coord namespace
from .angleunit import AngleUnit, arcsec, arcmin, degrees, hours, radians
from .angle import Angle, _Angle
from .celestial import CelestialCoord, _CelestialCoord

# This isn't imported to the coord namespace.  You need to do coord.util.blah
from . import util



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function

try:
    import cPickle as pickle
except ImportError:
    import pickle
import copy
from collections import Hashable, Counter
import functools
import time
import coord

# This file has some helper functions that are used by tests from multiple files to help
# avoid code duplication.

# The functions in this file are based heavily on the corresponding functions in the
# galsim_test_helpers in the GalSim/tests directory.

def do_pickle(obj, func=(lambda x : x), copyable=True, reprable=True):
    """Check that the object is picklable.  Also that it has basic == and != functionality.

    :param obj:         The object to be checked
    :param func:        If desired, some function of the object to use for equality checking.
                        [default: (lambda x : x)]
    :param copyable:    Whether the object should be expected to be equal after copy(obj).
                        e.g. things with an internal random state may not be equal after copying.
                        [default: True]
    :param reprable:    Whether the object should be expected to survive a round trip through
                        obj = eval(repr(obj)).  This is often a desirable property of objects,
                        but it may not always be possible (or efficient). [default: True]
    """
    # Test pickle round-trip returns an equivalent (but not identical) object
    print('Try pickling ',obj)  # Note: this implicitly checks that str(obj) works.
    obj2 = pickle.loads(pickle.dumps(obj))
    assert obj2 is not obj, "Obj %s not identical after pickling"%obj
    f1 = func(obj)
    f2 = func(obj2)
    assert f2 == f1, "func(obj) = %r\nfunc(obj2) = %r"%(f1, f2)

    # Test the hash values are equal for two equivalent objects.
    if isinstance(obj, Hashable):
        assert hash(obj) == hash(obj2), "Objects %s and %s have different hashes"%(obj,obj2)

    # Test that copy makes an equivalent copy.
    obj3 = copy.copy(obj)
    assert obj3 is not obj, "Obj %s not identical after copy"%obj
    if copyable:
        f1 = func(obj)  # Might need to redo this in case function changes an internal state.
        f3 = func(obj3)
        assert f3 == f1, "func(obj) = %r\nfunc(obj3) = %r"%(f1, f3)

    # A deepcopy should always work.
    obj4 = copy.deepcopy(obj)
    assert obj4 is not obj, "Obj %s not identical after deep copy"%obj
    f1 = func(obj)
    f4 = func(obj4)
    assert f4 == f1, "func(obj) = %r\nfunc(obj4) = %r"%(f1, f4)

    # Also test that the repr is an accurate representation of the object.
    # The gold standard is that eval(repr(obj)) == obj.  So check that here as well.

    if reprable:
        obj5 = eval(repr(obj))
        f1 = func(obj)
        f5 = func(obj5)
        assert f5 == f1, "func(obj) = %r\nfunc(obj5) = %r"%(f1, f5)
    else:
        # Even if we're not actually doing the test, still make the repr to check for syntax errors.
        repr(obj)


def all_obj_diff(objs):
    """Helper function that verifies that each element in `objs` is unique and, if hashable,
    produces a unique hash.
    """
    # Check that all objects are unique.
    # Would like to use `assert len(objs) == len(set(objs))` here, but this requires that the
    # elements of objs are hashable (and that they have unique hashes!, which is what we're trying
    # to test!.  So instead, we just loop over all combinations.
    for i, obji in enumerate(objs):
        # Could probably start the next loop at `i+1`, but we start at 0 for completeness
        # (and to verify a != b implies b != a)
        for j, objj in enumerate(objs):
            if i == j:
                continue
            assert obji != objj, ("Found equivalent objects {0} == {1} at indices {2} and {3}"
                                  .format(obji, objj, i, j))

    # Now check that all hashes are unique (if the items are hashable).
    if not isinstance(objs[0], Hashable):
        return
    hashes = [hash(obj) for obj in objs]
    try:
        assert len(hashes) == len(set(hashes)), "At least two of the hashes are equal"
    except AssertionError as e:
        for k, v in Counter(hashes).items():
            if v <= 1:
                continue
            print("Found multiple equivalent object hashes:")
            for i, obj in enumerate(objs):
                if hash(obj) == k:
                    print(i, repr(obj))
        raise e


def timer(f):
    """A decorator to put before each test function to have it output how long it took to run the
    test.
    """
    @functools.wraps(f)
    def f2(*args, **kwargs):
        t0 = time.time()
        result = f(*args, **kwargs)
        t1 = time.time()
        fname = repr(f).split()[1]
        print('time for %s = %.2f' % (fname, t1-t0))
        return result
    return f2



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
import numpy as np
import math
import coord

from helper_util import *

# We'll use this a lot, so just import it.
from math import pi

@timer
def test_init():
    """Test basic construction and use of Angle classes
    """
    # Various ways to make 45 degrees
    theta1 = pi/4. * coord.radians
    theta2 = 45 * coord.degrees
    theta3 = coord.Angle(3, coord.hours)
    theta4 = coord.Angle(theta=45 * 60, unit=coord.arcmin)
    theta5 = coord.Angle(45. * 3600, unit=coord.arcsec)
    gradians = coord.AngleUnit(2. * pi / 400.)
    theta6 = 50 * gradians
    theta7 = coord.Angle(theta1)  # Copy constructor

    # Equivalent ways to access theta in radians
    assert theta1.rad == pi/4., "(pi/4 radians).rad != pi/4"
    assert theta1 / coord.radians == pi/4., "(pi/4 radians) / radians != pi/4"

    # Others involved a calculation so aren't required to be precisely equal.
    np.testing.assert_almost_equal(theta2.rad, pi/4., decimal=12, err_msg='45 degrees != pi/4 rad')
    np.testing.assert_almost_equal(theta3.rad, pi/4., decimal=12, err_msg='3 hours != pi/4 rad')
    np.testing.assert_almost_equal(theta4.rad, pi/4., decimal=12, err_msg='45*60 arcmin != pi/4 rad')
    np.testing.assert_almost_equal(theta5.rad, pi/4., decimal=12, err_msg='45*3600 arcsec != pi/4 rad')
    np.testing.assert_almost_equal(theta6.rad, pi/4., decimal=12, err_msg='50 grad != pi/4 rad')
    np.testing.assert_almost_equal(theta7.rad, pi/4., decimal=12, err_msg='copy != pi/4 rad')

    # Access angle in other units
    np.testing.assert_almost_equal(theta1 / gradians, 50., 12)
    np.testing.assert_almost_equal(theta3 / gradians, 50., 12)
    np.testing.assert_almost_equal(theta1 / coord.hours, 3., 12)
    np.testing.assert_almost_equal(theta5 / coord.hours, 3., 12)


@timer
def test_invalid():
    """Test invalid ways to try to make an Angle
    """
    theta = 1.3 * coord.degrees

    # No unit
    np.testing.assert_raises(TypeError, coord.Angle, 3.4)

    # Invalid value type
    np.testing.assert_raises(TypeError, coord.Angle, theta, coord.degrees)
    np.testing.assert_raises(ValueError, coord.Angle, 'spam', coord.degrees)

    # Wrong order
    np.testing.assert_raises(TypeError, coord.Angle, coord.degrees, 1.3)

    # Invalid unit type
    np.testing.assert_raises(TypeError, coord.Angle, 1.3, 3)
    np.testing.assert_raises(TypeError, coord.Angle, 1.3, theta)

    # Invalid value type for copy constructor
    np.testing.assert_raises(TypeError, coord.Angle, 1.3)


@timer
def test_arith():
    """Test various arithmetic operations on Angles
    """
    theta1 = 45. * coord.degrees
    theta2 = 135. * coord.degrees
    np.testing.assert_almost_equal((theta1 + theta2).rad, pi, decimal=12)
    np.testing.assert_almost_equal((theta2 - theta1).rad, pi/2., decimal=12)
    np.testing.assert_almost_equal((4*theta1).rad, pi, decimal=12)
    np.testing.assert_almost_equal((theta2*2).rad, 1.5*pi, decimal=12)
    np.testing.assert_almost_equal((4*theta1 - theta2).rad, pi/4., decimal=12)
    np.testing.assert_almost_equal((theta1/2.).rad, pi/8., decimal=12)
    np.testing.assert_almost_equal((-theta1).rad, -pi/4., decimal=12)

    theta2 += theta1
    np.testing.assert_almost_equal(theta2.rad, pi, decimal=12)
    theta2 -= 2*theta1
    np.testing.assert_almost_equal(theta2.rad, pi/2., decimal=12)
    theta2 *= 4
    np.testing.assert_almost_equal(theta2.rad, 2.*pi, decimal=12)
    theta2 /= -2
    np.testing.assert_almost_equal(theta2.rad, -pi, decimal=12)

    # Check invalid arithmetic
    np.testing.assert_raises(TypeError, coord.Angle.__add__, theta1, 23)
    np.testing.assert_raises(TypeError, coord.Angle.__add__, theta1, '23')
    np.testing.assert_raises(TypeError, coord.Angle.__add__, theta1, coord.degrees)
    np.testing.assert_raises(TypeError, coord.Angle.__sub__, theta1, 23)
    np.testing.assert_raises(TypeError, coord.Angle.__sub__, theta1, '23')
    np.testing.assert_raises(TypeError, coord.Angle.__sub__, theta1, coord.degrees)
    np.testing.assert_raises(TypeError, coord.Angle.__mul__, theta1, '23')
    np.testing.assert_raises(TypeError, coord.Angle.__mul__, theta1, 23*coord.degrees)
    np.testing.assert_raises(TypeError, coord.Angle.__mul__, theta1, coord.degrees)
    np.testing.assert_raises(TypeError, coord.Angle.__div__, theta1, 23*coord.degrees)
    np.testing.assert_raises(TypeError, coord.Angle.__div__, theta1, '23')

@timer
def test_wrap():
    """Test the Angle.wrap function
    """
    # Two values that differ by a full cycle
    theta1 = 45 * coord.degrees
    theta2 = (45 + 360) * coord.degrees

    # Not wrapped intrinsically.  They differ by 2pi
    assert 6.28 < theta2.rad - theta1.rad < 6.29

    # Wrapping theta2 makes them equal
    print('theta1 = ',theta1,' => ',theta1.wrap())
    print('theta2 = ',theta2,' => ',theta2.wrap())
    np.testing.assert_almost_equal(theta2.wrap().rad, theta1.rad, decimal=12)

    # Any multiple of 360 degrees is equivalent
    theta3 = (45 - 360) * coord.degrees
    assert -6.29 < theta3.rad - theta1.rad < -6.28
    np.testing.assert_almost_equal(theta3.wrap().rad, theta1.rad, decimal=12)

    theta4 = (45 + 10 * 360) * coord.degrees
    assert 62.8 < theta4.rad - theta1.rad < 62.9
    np.testing.assert_almost_equal(theta4.wrap().rad, theta1.rad, decimal=12)

    # Can wrap around other angles besides 0.
    pir = pi * coord.radians
    np.testing.assert_almost_equal(theta2.wrap(pir).rad, theta1.rad, decimal=12)
    np.testing.assert_almost_equal(theta2.rad, theta1.wrap(2*pir).rad, decimal=12)
    np.testing.assert_almost_equal(theta2.rad, theta1.wrap(3*pir).rad, decimal=12)
    np.testing.assert_almost_equal(theta3.rad, theta1.wrap(-pir).rad, decimal=12)
    np.testing.assert_almost_equal(theta3.rad, theta1.wrap(-2*pir).rad, decimal=12)
    np.testing.assert_almost_equal(theta2.wrap(27 * coord.radians).rad,
                                   theta1.wrap(27 * coord.radians).rad, decimal=12)
    np.testing.assert_almost_equal(theta3.wrap(-127 * coord.radians).rad,
                                   theta1.wrap(-127 * coord.radians).rad, decimal=12)
    np.testing.assert_almost_equal(theta4.wrap(-927 * coord.radians).rad,
                                   theta1.wrap(-927 * coord.radians).rad, decimal=12)


@timer
def test_trig():
    """Test the various trig functions with Angles
    """
    theta1 = 45 * coord.degrees
    np.testing.assert_almost_equal(theta1.sin(), math.sqrt(0.5), decimal=12)
    np.testing.assert_almost_equal(theta1.cos(), math.sqrt(0.5), decimal=12)
    np.testing.assert_almost_equal(theta1.tan(), 1., decimal=12)
    np.testing.assert_almost_equal(theta1.sincos(), math.sqrt(0.5), decimal=12)

    theta2 = 30 * coord.degrees
    np.testing.assert_almost_equal(theta2.sin(), 0.5, decimal=12)
    np.testing.assert_almost_equal(theta2.cos(), math.sqrt(0.75), decimal=12)
    np.testing.assert_almost_equal(theta2.tan(), math.sqrt(1./3.), decimal=12)
    np.testing.assert_almost_equal(theta2.sincos(), (0.5, math.sqrt(0.75)), decimal=12)

    theta3 = 59 * coord.degrees
    theta4 = 31 * coord.degrees
    np.testing.assert_almost_equal(theta3.sin(), theta4.cos(), decimal=12)
    np.testing.assert_almost_equal(theta3.cos(), theta4.sin(), decimal=12)
    np.testing.assert_almost_equal(theta3.tan(), 1./theta4.tan(), decimal=12)
    np.testing.assert_almost_equal(theta3.sincos(), theta4.sincos()[::-1], decimal=12)

@timer
def test_pickle():
    """Test that Angle instances pickle correctly
    """
    theta1 = 45 * coord.degrees
    theta2 = 29.9 * coord.radians
    theta3 = 1.e-4 * coord.arcsec
    do_pickle(theta1)
    do_pickle(theta2)
    do_pickle(theta3)

@timer
def test_eq():
    """Check that equal angles are equal, but unequal ones are not.
    """
    theta1 = pi/4. * coord.radians
    theta2 = 45 * coord.degrees
    assert theta1 == theta2, "pi/4 rad != 45 deg"

    theta3 = coord.Angle(theta1)  # Copy constructor
    assert theta3 == theta1, "copy not == to original"

    theta4 = theta1.wrap()
    assert theta1 == theta4, "(pi/4 rad).wrap() != pi/4 rad"

    # These should all test as unequal.  Note some non-Angles in the list.
    diff_list = [ theta1,
                  14 * coord.degrees,
                  -theta1,
                  theta1 * 2.,
                  theta1 + 360. * coord.degrees,
                  theta1 - 360. * coord.degrees,
                  pi/4.,
                  coord.Angle,
                  None ]
    all_obj_diff(diff_list)

@timer
def test_hms():
    """Test Angle.hms and from_hms functions
    """
    theta1 = 45 * coord.degrees
    hms1 = theta1.hms()
    print('hms1 = ',hms1)
    assert hms1 == '03:00:00'

    theta2 = 0 * coord.degrees
    hms2 = theta2.hms(plus_sign=True, prec=8, sep='hm')
    print('hms2 = ',hms2)
    assert hms2 == '+00h00m00.00000000'

    # Check that 360 degres is 24 hours
    theta3 = 360 * coord.degrees
    hms3 = theta3.hms(sep='hms', prec=0)
    print('hms3 = ',hms3)
    assert hms3 == '24h00m00s'

    # Check a negative small value
    theta4 = -0.2 * coord.degrees
    hms4 = theta4.hms(pad=False, prec=2, sep=(' hours ',' minutes ',' seconds'))
    print('hms4 = ',hms4)
    assert hms4 == '-0 hours 0 minutes 48.00 seconds'

    # Some random value
    theta5 = 1.3 * coord.radians
    hms5 = theta5.hms(prec=12)
    print('hms5 = ',hms5)
    h,m,s = hms5.split(':')
    theta5b = float(h) + float(m)/60. + float(s)/3600.
    np.testing.assert_almost_equal(theta5 / coord.hours, theta5b, decimal=12)

    # Negative angles < 1 degree are tricky
    theta6 = -6.112479 * coord.arcmin
    hms6 = theta6.hms(prec=2)
    print('hms6 = ',hms6)
    assert hms6 == '-00:00:24.45'
    alt_theta6 = -6.1125 * coord.arcmin

    # Check round trip
    np.testing.assert_almost_equal(theta1.rad, coord.Angle.from_hms(hms1).rad, decimal=12)
    np.testing.assert_almost_equal(theta2.rad, coord.Angle.from_hms(hms2).rad, decimal=12)
    np.testing.assert_almost_equal(theta3.rad, coord.Angle.from_hms(hms3).rad, decimal=12)
    np.testing.assert_almost_equal(theta4.rad, coord.Angle.from_hms(hms4).rad, decimal=12)
    np.testing.assert_almost_equal(theta5.rad, coord.Angle.from_hms(hms5).rad, decimal=12)
    np.testing.assert_almost_equal(alt_theta6.rad, coord.Angle.from_hms(hms6).rad, decimal=12)

    # This one doesn't round trip, but it is valid -> hms
    theta7 = 1.2345 * coord.radians
    hms7 = theta7.hms(sep='', prec=0)
    print('hms7 = ',hms7)
    assert hms7 == '044256'

    # These aren't constructible from hms() but they parse correctly
    theta8 = 17.3 * coord.hours
    hms8a = '17h18'
    hms8b = '+17h18m'
    hms8c = '17.3h'
    hms8d = '+17.1h12m'  # Weird, but allowed.  Maybe it shouldn't be.
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_hms(hms8a).rad, decimal=12)
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_hms(hms8b).rad, decimal=12)
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_hms(hms8c).rad, decimal=12)
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_hms(hms8d).rad, decimal=12)

    # Check invalid calls
    np.testing.assert_raises(ValueError, theta1.hms, sep='hmsss')
    np.testing.assert_raises(TypeError, theta1.hms, sep=coord.hours)
    np.testing.assert_raises(ValueError, theta1.hms, prec=-1)
    np.testing.assert_raises(TypeError, theta1.hms, prec='lots')
    np.testing.assert_raises(ValueError, coord.Angle.from_hms, 'a')
    np.testing.assert_raises(ValueError, coord.Angle.from_hms, '-')
    np.testing.assert_raises(ValueError, coord.Angle.from_hms, '+')
    np.testing.assert_raises(ValueError, coord.Angle.from_dms, '-17')
    np.testing.assert_raises(ValueError, coord.Angle.from_hms, '')
    np.testing.assert_raises(ValueError, coord.Angle.from_hms, '01:21:31:15')



@timer
def test_dms():
    """Test Angle.dms and from_dms functions
    """
    theta1 = 45 * coord.degrees
    dms1 = theta1.dms()
    print('dms1 = ',dms1)
    assert dms1 == '45:00:00'

    theta2 = 0 * coord.degrees
    dms2 = theta2.dms(plus_sign=True, prec=8, sep='dm')
    print('dms2 = ',dms2)
    assert dms2 == '+00d00m00.00000000'

    theta3 = 360 * coord.degrees
    dms3 = theta3.dms(sep='dms', prec=0)
    print('dms3 = ',dms3)
    assert dms3 == '360d00m00s'

    theta4 = -1 * coord.degrees
    dms4 = theta4.dms()
    dms4 = theta4.dms(pad=False, prec=2, sep=(' degrees ',' minutes ',' seconds'))
    print('dms4 = ',dms4)
    assert dms4 == '-1 degrees 0 minutes 0.00 seconds'

    theta5 = 1.3 * coord.radians
    dms5 = theta5.dms(prec=12)
    print('dms5 = ',dms5)
    d,m,s = dms5.split(':')
    theta5b = float(d) + float(m)/60. + float(s)/3600.
    np.testing.assert_almost_equal(theta5 / coord.degrees, theta5b, decimal=12)

    # Negative angles < 1 degree are tricky
    theta6 = -6.112479 * coord.arcmin
    dms6 = theta6.dms(prec=2)
    print('dms6 = ',dms6)
    assert dms6 == '-00:06:06.75'
    alt_theta6 = -6.1125 * coord.arcmin

    # Check round trip
    np.testing.assert_almost_equal(theta1.rad, coord.Angle.from_dms(dms1).rad, decimal=12)
    np.testing.assert_almost_equal(theta2.rad, coord.Angle.from_dms(dms2).rad, decimal=12)
    np.testing.assert_almost_equal(theta3.rad, coord.Angle.from_dms(dms3).rad, decimal=12)
    np.testing.assert_almost_equal(theta4.rad, coord.Angle.from_dms(dms4).rad, decimal=12)
    np.testing.assert_almost_equal(theta5.rad, coord.Angle.from_dms(dms5).rad, decimal=12)
    np.testing.assert_almost_equal(alt_theta6.rad, coord.Angle.from_dms(dms6).rad, decimal=12)

    # This one doesn't round trip, but it is valid -> hms
    theta7 = 1.2345 * coord.radians
    dms7 = theta7.dms(sep='', prec=0)
    print('dms7 = ',dms7)
    assert dms7 == '704354'

    # These aren't constructible from dms() but they parse correctly
    theta8 = 17.3 * coord.degrees
    dms8a = '+17d18'
    dms8b = '17d18m'
    dms8c = '+17.3d'
    dms8d = '17.1d12m'  # Weird, but allowed.  Maybe it shouldn't be.
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_dms(dms8a).rad, decimal=12)
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_dms(dms8b).rad, decimal=12)
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_dms(dms8c).rad, decimal=12)
    np.testing.assert_almost_equal(theta8.rad, coord.Angle.from_dms(dms8d).rad, decimal=12)

    # Check invalid calls
    np.testing.assert_raises(ValueError, theta1.dms, sep='dmsss')
    np.testing.assert_raises(TypeError, theta1.dms, sep=coord.degrees)
    np.testing.assert_raises(ValueError, theta1.dms, prec=-1)
    np.testing.assert_raises(TypeError, theta1.dms, prec='lots')
    np.testing.assert_raises(ValueError, coord.Angle.from_dms, 'a')
    np.testing.assert_raises(ValueError, coord.Angle.from_dms, '-')
    np.testing.assert_raises(ValueError, coord.Angle.from_dms, '+')
    np.testing.assert_raises(ValueError, coord.Angle.from_dms, '+17')
    np.testing.assert_raises(ValueError, coord.Angle.from_dms, '')
    np.testing.assert_raises(ValueError, coord.Angle.from_dms, '01:21:31:15')


if __name__ == '__main__':
    test_init()
    test_invalid()
    test_arith()
    test_wrap()
    test_trig()
    test_pickle()
    test_eq()
    test_hms()
    test_dms()



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
import numpy as np
import math
import coord
import datetime

from helper_util import *

# Note: sun_position_ecliptic and ecliptic_obliquity in util.py are implicitly tested
# by the test_ecliptic_date tests.

@timer
def test_julian():
    """Test date_to_julian_day function
    """
    # Use Julian Day Number from Date Calculator here:
    #    http://keisan.casio.com/exec/system/1227779487
    # The Julian Day for noon on 1/1/2000 is also mentioned here:
    #    https://en.wikipedia.org/wiki/Julian_day
    dates = [ datetime.date(1582, 10, 14),
              datetime.date(2000, 1, 1),
              datetime.datetime(2000, 1, 1, 12, 0, 0),
              datetime.datetime(2000, 1, 1, 0, 0, 0),
              datetime.datetime(2000, 1, 1, 18, 0, 0)
            ]
    julian_days = [ 2299160, 2451545, 2451545, 2451544.5, 2451545.25 ]

    for date, jd in zip(dates, julian_days):
        jd1 = coord.util.date_to_julian_day(date)
        np.testing.assert_almost_equal(jd1, jd, 6, 'Julian day not as expected for %s'%date)

    np.testing.assert_raises(ValueError, coord.util.date_to_julian_day, 2000)
    np.testing.assert_raises(ValueError, coord.util.date_to_julian_day, '1875 AD')



if __name__ == '__main__':
    test_julian()



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
import math

class AngleUnit(object):
    """
    A class for defining angular units used by Angle objects.

    **Initialization:**

        An AngleUnit takes a single argument for initialization, a float that specifies the size
        of the desired angular unit in radians.  For example:

            :meth:`coord.AngleUnit.__init__`
 
            >>> gradian = AngleUnit(2. * math.pi / 400.)
            >>> print(gradian)
            coord.AngleUnit(0.015707963267948967)

    **Built-in units:**

        There are five built-in AngleUnits which are always available for use:

            :coord.radians:   coord.AngleUnit(1.)
            :coord.degrees:   coord.AngleUnit(pi / 180.)
            :coord.hours:     coord.AngleUnit(pi / 12.)
            :coord.arcmin:    coord.AngleUnit(pi / 180. / 60.)
            :coord.arcsec:    coord.AngleUnit(pi / 180. / 3600.)

    **Attribute:**

        An AngleUnit as the following (read-only) attribute:

            :value:     The measure of the unit in radians.
    """
    valid_names = ['rad', 'deg', 'hr', 'hour', 'arcmin', 'arcsec']

    def __init__(self, value):
        """
        :param value:   The measure of the unit in radians.
        """
        self._value = float(value)

    @property
    def value(self):
        """A read-only attribute giving the measure of the AngleUnit in radians."""
        return self._value

    def __rmul__(self, theta):
        """float * AngleUnit returns an Angle"""
        from .angle import Angle  # Can't do this at the top, since circular reference.
        return Angle(theta, self)

    def __div__(self, unit):
        """AngleUnit / AngleUnit returns a float giving the relative scaling.

        Note: At least to within machine precision, it is the case that

            (x * angle_unit1) / angle_unit2 == x * (angle_unit1 / angle_unit2)
        """
        if not isinstance(unit, AngleUnit):
            raise TypeError("Cannot divide AngleUnit by %s"%unit)
        return self.value / unit.value

    __truediv__ = __div__

    @staticmethod
    def from_name(unit):
        """Convert a string into the corresponding AngleUnit.

        Only the start of the string is checked, so for instance 'radian' or 'radians' is
        equivalent to 'rad'.

        Valid options are:

            :rad:           AngleUnit(1.)
            :deg:           AngleUnit(pi / 180.)
            :hour or hr:    AngleUnit(pi / 12.)
            :arcmin:        AngleUnit(pi / 180. / 60.)
            :arcsec:        AngleUnit(pi / 180. / 3600.)

        Note: these valid names are listed in AngleUnit.valid_names.

        :param unit:    The string name of the unit to return

        :returns: an AngleUnit
        """
        unit = unit.strip().lower()
        if unit.startswith('rad') :
            return radians
        elif unit.startswith('deg') :
            return degrees
        elif unit.startswith('hour') :
            return hours
        elif unit.startswith('hr') :
            return hours
        elif unit.startswith('arcmin') :
            return arcmin
        elif unit.startswith('arcsec') :
            return arcsec
        else :
            raise ValueError("Unknown Angle unit: %s"%unit)

    def __repr__(self):
        if self == radians:
            return 'coord.radians'
        elif self == degrees:
            return 'coord.degrees'
        elif self == hours:
            return 'coord.hours'
        elif self == arcmin:
            return 'coord.arcmin'
        elif self == arcsec:
            return 'coord.arcsec'
        else:
            return 'coord.AngleUnit(%r)'%self.value

    def __eq__(self, other):
        return isinstance(other,AngleUnit) and self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(('coord.AngleUnit', self.value))

# Convenient pre-set built-in units
# (These are typically the only ones we will use.)

radians = AngleUnit(1.)
hours = AngleUnit(math.pi / 12.)
degrees = AngleUnit(math.pi / 180.)
arcmin = AngleUnit(math.pi / 10800.)
arcsec = AngleUnit(math.pi / 648000.)



from __future__ import print_function
import sys,os,glob,re

import setuptools
from setuptools import setup, Extension

print('Python version = ',sys.version)
py_version = "%d.%d"%sys.version_info[0:2]  # we check things based on the major.minor version.

with open('requirements.txt') as f:
    required = f.read().splitlines()

sources = glob.glob(os.path.join('src','*.cpp'))
print('sources = ',sources)
headers = glob.glob(os.path.join('include','*.h'))
print('headers = ',headers)
ext=Extension("coord._coord", sources, depends=headers, include_dirs=['include'])

with open('README.rst') as file:
    long_description = file.read()

# Read in the coord version from coord/_version.py
# cf. http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
version_file=os.path.join('coord','_version.py')
verstrline = open(version_file, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    coord_version = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (version_file,))
print('Coord version is %s'%(coord_version))

dist = setup(name="LSSTDESC.Coord",
      version=coord_version,
      author="LSST DESC (contact: Mike Jarvis)",
      author_email="michael@jarvis.net",
      description="Python module for handling angles and celestial coordinates",
      long_description=long_description,
      license = "MIT License",
      url="https://github.com/LSSTDESC/Coord",
      download_url="https://github.com/LSSTDESC/Coord/releases/tag/v%s.zip"%coord_version,
      packages=['coord'],
      package_data={'coord' : headers },
      ext_modules=[ext],
      install_requires=required)

# setup.py doesn't put the .so file in the coord directory, so this bit makes it possible to
# import coord from the root directory.  Not really advisable, but everyone does it at some
# point, so might as well facilitate it.
build_lib = glob.glob(os.path.join('build','*','coord','_coord*.so'))
if len(build_lib) >= 1:
    lib = os.path.join('coord','_coord.so')
    if os.path.lexists(lib): os.unlink(lib)
    os.link(build_lib[0], lib)

